/*-
 * wrapper library for long paths exceeding PATH_MAX
 * 
 * Copyright (c) 2013 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Nils Goroll <nils.goroll@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * == IMPLEMENTATION NOTES ==
 *
 * UNIMPLEMENTED
 * - ftw, nftw
 * - copylist
 * - stdio.h stuff: fopen, freopen
 * - acl
 * - Check for O_SEARCH ?
 *
 * TODO
 * - use pathconf instead of fixed MAX_PATH
 * - cache intermediate fds
 * - cache getpath results
 *
 * Notes on lf64:
 * for the intermediate filedescriptors, we always use openat64 unless _LP64 we
 * can do so because we never hand out these fds - Xatl/Xatl2 always close them
 *
 * the lib gets compiled with _FILE_OFFSET_BITS==32 for 32 bit builds so we can
 * provide both variants. longpath.h will provide the appropriate redefinitions
 * depending on our consumer
 *
 * Notes on PATH_MAX:
 * Ref: http://pubs.opengroup.org/onlinepubs/009695399/basedefs/limits.h.html
 * Maximum number of bytes in a pathname, including the terminating null character.
 *
 */

#include "../config.h"

#define _ATFILE_SOURCE

#if defined(_LP64)
#define	_FILE_OFFSET_BITS	64
#else
#define	_FILE_OFFSET_BITS	32
#endif
#define	_LARGEFILE_SOURCE
#define	_LARGEFILE64_SOURCE


#ifdef HAVE_ASSERT_H
#include <assert.h>
#else
#define assert(EX) ((void)0)
#endif

/*
 * Thanks to Jonathan Wakely for these macros
 * http://gcc.gnu.org/ml/gcc-help/2011-01/msg00121.html
 *
 */
#if ((__GNUC__ * 100) + __GNUC_MINOR__) >= 405
# define GCC_DIAG_DO_PRAGMA(x) _Pragma (#x)
# define GCC_DIAG_PRAGMA(x) GCC_DIAG_DO_PRAGMA(GCC diagnostic x)
# if ((__GNUC__ * 100) + __GNUC_MINOR__) >= 406
#  define GCC_DIAG_OFF(x) GCC_DIAG_PRAGMA(push) \
     GCC_DIAG_PRAGMA(ignored x)
#  define GCC_DIAG_ON(x) GCC_DIAG_PRAGMA(pop)
# else
#  define GCC_DIAG_OFF(x) GCC_DIAG_PRAGMA(ignored x)
#  define GCC_DIAG_ON(x)  GCC_DIAG_PRAGMA(warning x)
# endif
#else
# define GCC_DIAG_OFF(x)
# define GCC_DIAG_ON(x)
#endif

/* 
 * work around solaris implementing stat() and lstat() in sys/stat_impl.h
 * on the basis of _xstat()
 */
#ifdef INTERPOSE
# ifdef HAVE__XSTAT
#  ifdef HAVE_SYS_STAT_IMPL_H
#   define _SYS_STAT_IMPL_H
#include <fcntl.h>
extern int mknod(const char *, mode_t, dev_t);
extern int stat(const char *, struct stat *);
extern int lstat(const char *, struct stat *);
extern int fstatat(int, const char *, struct stat *, int);
#  else
#   error This machine probably needs different _xstat handling
#  endif
# endif /* HAVE__XSTAT */

# include <longpath/longpath_common.h>

# define real(x)		real_ ## x
# if defined(_LP64)
#  define real64(x)		real_ ## x
# else
#  define real64(x)		real_ ## x	## 64
# endif
# define our(x)				 x
# if defined(_LP64)
#  define our64(x)			 x
# else
#  define our64(x)			 x	## 64
# endif

/*
 * declaration of real() function pointers
 */
# define INTERPOSE_REAL
# define FUNCTION(type, name, ...) static type (* real(name))(__VA_ARGS__);
# if ! defined(_LP64)
#  define FUNCTION64(type, name, ...) static type (* real64(name))(__VA_ARGS__);
# else
#  define FUNCTION64(type, name, ...)
# endif /* _LP64 */
# include <longpath/longpath_tbl_functions.h>
# include <longpath/longpath_tbl_functions64.h>
# undef FUNCTION
# undef FUNCTION64
# undef INTERPOSE_REAL

# include <dlfcn.h>
# include <stdio.h>
# define init_dlsym(fptr, name) do {					\
		if (! (fptr = dlsym(RTLD_NEXT, #name))) {		\
			perror("looking up symbol " #name " failed");	\
			exit(1);					\
		}							\
	} while(0)

#ifdef TRACE
static int tracefd;
static pid_t pid;
//static void *my_dl_handle;

#include <unistd.h>
#endif

/*
 * init real() function ponters
 */
void __attribute__ ((constructor)) longpath_init(void);

GCC_DIAG_OFF("-pedantic")

void __attribute__ ((constructor))
longpath_init(void) {
# define FUNCTION(type, name, ...) init_dlsym(real(name), name);
# if ! defined(_LP64)
#  define FUNCTION64(type, name, ...) init_dlsym(real64(name), name ## 64);
# else
#  define FUNCTION64(type, name, ...)
# endif /* _LP64 */
# include <longpath/longpath_tbl_functions.h>
# include <longpath/longpath_tbl_functions64.h>
# undef FUNCTION
# undef FUNCTION64

#ifdef TRACE
	tracefd = real(open)("/tmp/lptrace", O_WRONLY|O_CREAT|O_APPEND, 0600);
	if (tracefd == -1) {
		perror("trace file ");
		exit(1);
	}
	pid = getpid();
#endif /* TRACE */
}

GCC_DIAG_ON("-pedantic")

#ifdef TRACE
#define TRACELEN 1024
#define trace(fmt, ...) do {						\
		char _buf[TRACELEN];					\
		int _l = snprintf(_buf, TRACELEN, "%d: " fmt, pid, ##__VA_ARGS__); \
		if ((_l > 0) && (_l < TRACELEN))			\
			(void)write(tracefd, _buf, _l);			\
	} while(0)

#include <stdio.h>
#include <string.h>
void *
our(dlopen)(const char *pathname, int mode) {
	trace("dlopen(%s, 0x%x)\n", pathname ? pathname : "(null)", mode);
	if (pathname && strstr(pathname, "libc.so")) {
//		const int nmode = RTLD_GROUP|RTLD_FIRST;
		trace("returning dlopen(%s, 0x%x)\n", OWN_PATH, mode);
		return real(dlopen)(OWN_PATH, mode);
	}
	if (mode & RTLD_GLOBAL) {
		const int nmode = RTLD_NOW|RTLD_GLOBAL;
		void *r;
		trace("actual dlopen(%s, 0x%x)\n", OWN_PATH, mode|RTLD_PARENT);
		r = real(dlopen)(pathname, mode|RTLD_PARENT);
		trace("doing additional dlopen(%s, 0x%x)\n", OWN_PATH, nmode);
		(void) real(dlopen)(OWN_PATH, nmode);
		return r;
	}
	return real(dlopen)(pathname, mode);
}
#endif /* TRACE */

#else /* ! INTERPOSE */
# include <longpath/longpath.h>
# define real(x)			 x
# if defined(_LP64)
#  define real64(x)			 x
# else
#  define real64(x)			 x	## 64
# endif
# define our(x)				 x ## l
# if defined(_LP64)
#  define our64(x)			 x ## l
# else
#  define our64(x)			 x ## l	## 64
# endif
#endif /* INTERPOSE */

// string functions
#if defined(HAVE_STRING_H)
# include <string.h>
#endif

#ifdef HAVE_LIMITS_H
# include <limits.h>
#endif
#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif
#ifdef HAVE_STDIO_H
# include <stdio.h>
#endif

// varargs for open
#if   defined(HAVE_SYS_VARARGS_H)
# include <sys/varargs.h>
#elif defined(HAVE_STDARG_H)
# include <stdarg.h>
#endif

// opendir
#ifdef HAVE_DIRENT_H
# include <dirent.h>
#endif

// linux xattr
#ifdef HAVE_FGETXATTR
# include <sys/xattr.h>
#endif

static int _openat(int, const char *, int, mode_t mode);
#if !defined(_LP64)
static int _openat64(int, const char *, int, mode_t mode);
#else
#define    _openat64(fd, p, f, m)	_openat(fd, p, f, m)
#endif

/*
 * close an intermediate file descriptor, restoring errno if necessary
 */
static inline void
close_ifd(int ifd) {
	const int s = errno;

	assert(ifd >= 0);
	assert(ifd != (int)AT_FDCWD);

	if (close(ifd) == -1)
		errno = s;
}

#ifdef USE_CHDIR_FALLBACK

/*
 * define NEED_CHDIR when using these
 */
static inline void restore_cwd(int *fd);
static int chdir_im(const char *path, const char **remaining, int limit);

#define Xl(funcn, path, ...)						\
	int r;								\
	int ocwd = real(open)(".", O_RDONLY|O_NDELAY|O_LARGEFILE); \
	const char *remaining;						\
									\
	if (ocwd == -1)							\
		return -1;						\
									\
	if (chdir_im(path, &remaining, 0) == -1) {			\
		restore_cwd(&ocwd);					\
		return -1;						\
	}								\
									\
	path += (remaining - path);					\
									\
	r = funcn(path, ##__VA_ARGS__);					\
									\
	restore_cwd(&ocwd);						\
	return r

#endif /* USE_CHDIR_FALLBACK */

/*
 * oatl_im = open at long intermediate
 *
 * for a path longer than PATH_MAX, open intermediate directories and return an
 * fd to the last directory component required for colling *at functions on the
 * remaining path component
 *
 * filedes can be AT_FDCWD
 * openatl_interm will not close filedes
 *
 * return:
 *
 *        -1	for errors, errno will be set
 * filedes	if path is NULL or shorter than PATH_MAX
 * otherwise	new open filedescriptor
 *
 * The caller must close the returned fd unless it is equal to filedes
 */

static int
oatl_im(int filedes, const char *path, const char **remaining) {
	char		interm[PATH_MAX];
	const char	*p, *sl;
	size_t		l;
	int		fd;

	/*
	 * work with functions on NULL paths like fchownat
	 * - at any rate we can leave error handling with the system function
	 */
	if (path == NULL) {
		*remaining = NULL;
		return filedes;
	}

	l = strlen(path);

	if (l < PATH_MAX) {
		*remaining = path;
		return filedes;
	}

	p = path;
	fd = filedes;
	while (l >= PATH_MAX) {
		int nfd;

		assert(l == strlen(p));

		/* find last slash for strlen(component) < PATH_MAX */

		sl = p + PATH_MAX - 1;

		/*
		 * skip all trailing slashes before starting the search, this is
		 * required because otherwise the 'rest' path could start with a
		 * slash (would work on wrong path) or be the empty string
		 * (yield undefined error)
		 */
		while ((sl > p) && (*sl == '/'))
			sl--;

		while (sl > p) {
			if (*sl == '/')
				goto found;
			sl--;
		}

		errno = ENAMETOOLONG;
		goto err;

	  found:
		assert((sl - p) < PATH_MAX);
		strncpy(interm, p, (sl - p));
		interm[sl - p] = '\0';

		nfd = real64(openat)(fd, interm,
		    O_RDONLY|O_NDELAY|O_LARGEFILE, 0);

		if (nfd == -1)
			goto err;

		if (fd != filedes)
			close_ifd(fd);

		fd = nfd;
		l -= (sl - p) + 1;
		p  = sl + 1;
	}

	assert(fd != filedes);
	*remaining = p;
	return fd;

    err:
	if (fd != filedes)
		close_ifd(fd);
	return -1;
}

// long path *at function
#define Xatl(funcn, filedes, path, ...)					\
	int		__ifd, __fd;					\
	const char	*__rest;					\
									\
	__ifd = oatl_im(filedes, path, &__rest);			\
									\
	if (__ifd == -1)						\
		return -1;						\
									\
	__fd = funcn(__ifd, __rest, ##__VA_ARGS__);			\
									\
	if (__ifd != filedes)						\
		close_ifd(__ifd);					\
									\
	return __fd

// long path *at function with two paths
#define Xatl2(funcn, fd1, p1, fd2, p2, ...)				\
	int		__ifd1, __ifd2, __fd;				\
	const char	*__rest1;					\
	const char	*__rest2;					\
									\
	__ifd1 = oatl_im(fd1, p1, &__rest1);				\
	if (__ifd1 == -1)						\
		return -1;						\
									\
	__ifd2 = oatl_im(fd2, p2, &__rest2);				\
	if (__ifd2 == -1) {						\
		__fd = -1;						\
		goto close_ifd1;					\
	}								\
									\
	__fd = funcn(__ifd1, __rest1, __ifd2, __rest2, ##__VA_ARGS__);	\
									\
	if (__ifd2 != fd2)						\
		close_ifd(__ifd2);					\
									\
  close_ifd1:								\
	if (__ifd1 != fd1)						\
		close_ifd(__ifd1);					\
									\
	return __fd

// long path f* function - make mode an argument?
#define Xfl(path, flags, err, r, funcn, ...)				\
	int		__ifd;						\
									\
	__ifd = _openat64(AT_FDCWD, (path),				\
	    O_RDONLY|O_NDELAY|O_LARGEFILE|(flags), 0);			\
									\
	if (__ifd == -1)						\
		return (err);						\
									\
	r = funcn(__ifd, ##__VA_ARGS__);				\
									\
	close_ifd(__ifd)

/* OPEN */
static int
_openat(int filedes, const char *path, int oflag, mode_t mode) {
	Xatl(real(openat), filedes, path, oflag, mode);
}

#if !defined(_LP64)
static int
_openat64(int filedes, const char *path, int oflag, mode_t mode) {
	Xatl(real64(openat), filedes, path, oflag, mode);
}
#endif

#define Openatl(bits)						\
	mode_t mode = 0;					\
	if (oflag & O_CREAT) {					\
		va_list arg;					\
		va_start(arg, oflag);				\
		mode = va_arg(arg, mode_t);			\
		va_end(arg);					\
	}							\
								\
	return _openat ## bits (filedes, path, oflag, mode)

int
our(openat)(int filedes, const char *path, int oflag, ...) {
	Openatl(/*32*/);
}

#if !defined(_LP64)
int
our64(openat)(int filedes, const char *path, int oflag, ...) {
	Openatl(64);
}
#endif

#define Openl(bits)						\
	mode_t mode = 0;					\
	if (oflag & O_CREAT) {					\
		va_list arg;					\
		va_start(arg, oflag);				\
		mode = va_arg(arg, mode_t);			\
		va_end(arg);					\
	}							\
								\
	return _openat ## bits (AT_FDCWD, path, oflag, mode)


int
our(open)(const char *path, int oflag, ...) {
	Openl(/*32*/);
}
#if !defined(_LP64)
int
our64(open)(const char *path, int oflag, ...) {
	Openl(64);
}
#endif

/* CREAT */

int
our(creat)(const char *path, mode_t mode) {
	return _openat(AT_FDCWD, path, O_WRONLY|O_CREAT|O_TRUNC, mode);
}
#if !defined(_LP64)
int
our64(creat)(const char *path, mode_t mode) {
	return _openat64(AT_FDCWD, path, O_WRONLY|O_CREAT|O_TRUNC, mode);
}
#endif

/* ATTROPEN */
#ifdef HAVE_ATTROPEN
#define Attropenl(bits)							\
        int fd, afd;							\
        va_list ap;							\
									\
	fd = our64(open) (path, O_RDONLY|O_NONBLOCK);			\
	if (fd == -1)							\
		return -1;						\
									\
	va_start(ap, oflag);						\
	afd = _openat ## bits (fd, apath, oflag|O_XATTR,		\
	    va_arg(ap, mode_t));					\
	va_end(ap);							\
									\
	close_ifd(fd);							\
	return afd

int
our(attropen)(const char *path, const char *apath, int oflag, ...) {
	Attropenl(/*32*/);
}
#if !defined(_LP64)
int
our64(attropen)(const char *path, const char *apath, int oflag, ...) {
	Attropenl(64);
}
#endif
#endif /* HAVE_ATTROPEN */

/* GETXATTR / LGETXATTR */
/* untested */
#ifdef HAVE_FGETXATTR

#ifndef HAVE_FSETXATTR
#error Assuming fsetxattr to exist when fgetxattr exists
#endif

ssize_t
our(getxattr)(const char *path, const char *name, void *value, size_t size) {
	ssize_t r;
	Xfl(path, 0, -1,
	    r, real(fgetxattr), name, value, size);
	return r;
}
ssize_t
our(lgetxattr)(const char *path, const char *name, void *value, size_t size) {
	ssize_t r;
	Xfl(path, AT_SYMLINK_NOFOLLOW, -1,
	    r, real(fgetxattr), name, value, size);
	return r;
}
int
our(setxattr)(const char *path, const char *name, const void *value, size_t size, int flags) {
	int r;
	Xfl(path, 0, -1,
	    r, real(fsetxattr), name, value, size, flags);
	return r;
}
int
our(lsetxattr)(const char *path, const char *name, const void *value, size_t size, int flags) {
	int r;
	Xfl(path, AT_SYMLINK_NOFOLLOW, -1,
	    r, real(fsetxattr), name, value, size, flags);
	return r;
}
#endif /* HAVE_FGETXATTR */


/* LINK */

#ifdef HAVE_LINKAT
int
our(linkat)(int fd1, const char *p1, int fd2, const char *p2, int flags) {
	Xatl2(real(linkat), fd1, p1, fd2, p2, flags);
}

int
our(link)(const char *p1, const char *p2) {
	return our(linkat)(AT_FDCWD, p1, AT_FDCWD, p2, 0);
}
#else /* HAVE_LINKAT */
#ifndef USE_CHDIR_FALLBACK
#warning linkat missing - link* calls on long paths may fail
int
our(link)(const char *p1, const char *p2) {
	return real(link)(p1, p2);
}

#else /* USE_CHDIR_FALLBACK */
#warning linkat missing - link* calls on long paths may fail (trying prefix match)
static int
strprefixlen(const char *s1, const char *s2) {
	int i = 0;
	while((*s1 != '\0') && (*s1 == *s2)) {
		s1++; s2++; i++;
	}

	return i;
}

int
our(link)(const char *p1, const char *p2) {
	int l = strprefixlen(p1, p2);

	if (l > 0) {
		int r;
		int ocwd = real(open)(".", O_RDONLY|O_NDELAY|O_LARGEFILE);
		const char *remaining;

		if (ocwd == -1)
			return -1;

		if (chdir_im(p1, &remaining, l) == -1) {
			restore_cwd(&ocwd);
			return -1;
		}

		l = remaining - p1;

		p1 += l;
		p2 += l;

		r = real(link)(p1, p2);

		restore_cwd(&ocwd);
		return r;
	}

	return real(link)(p1, p2);
}
#endif /* USE_CHDIR_FALLBACK */
#endif /* HAVE_LINKAT */

/** UNLINK **/

int
our(unlinkat)(int fd, const char *p, int flags) {
	Xatl(real(unlinkat), fd, p, flags);
}

int
our(unlink)(const char *p) {
	return our(unlinkat)(AT_FDCWD, p, 0);
}

/** SYMLINK **/
#ifdef HAVE_SYMLINKAT
/* slightly differnt signature - damn */
int
our(symlinkat)(const char *dest, int filedes, const char *path) {
	int		ifd, r;
	const char	*rest;

	ifd = oatl_im(filedes, path, &rest);

	if (ifd == -1)
		return -1;

	r = real(symlinkat)(dest, ifd, rest);

	if (ifd != filedes)
		close_ifd(ifd);

	return r;
}
int
our(symlink)(const char *dest, const char *path) {
	return our(symlinkat)(dest, AT_FDCWD, path);
}
#else /* HAVE_SYMLINKAT */
#ifndef USE_CHDIR_FALLBACK
#warning symlinkat missing - symlink* calls on long paths may fail
int
our(symlink)(const char *dest, const char *path) {
	return real(symlink)(dest, path);
}
#else /* USE_CHDIR_FALLBACK */
int
our(symlink)(const char *dest, const char *path) {
	int r;
	int ocwd = real(open)(".", O_RDONLY|O_NDELAY|O_LARGEFILE);
	const char *remaining;

	if (ocwd == -1)
		return -1;

	if (chdir_im(path, &remaining, 0) == -1) {
		restore_cwd(&ocwd);
		return -1;
	}

	path += (remaining - path);

	r = real(symlink)(dest, path);

	restore_cwd(&ocwd);
	return r;
}

#endif /* USE_CHDIR_FALLBACK */
#endif /* HAVE_SYMLINKAT */


/** READLINK **/

#ifdef HAVE_READLINKAT
ssize_t
our(readlinkat)(int fd, const char *name, char *buf, size_t count) {
	Xatl(real(readlinkat), fd, name, buf, count);
}

ssize_t
our(readlink)(const char *name, char *buf, size_t count) {
	return our(readlinkat)(AT_FDCWD, name, buf, count);
}
#else /* HAVE_READLINKAT */
#ifndef USE_CHDIR_FALLBACK
#warning readlinkat missing - link* calls on long paths may fail
ssize_t
our(readlink)(const char *name, char *buf, size_t count) {
	return real(readlink)(name, buf, count);
}
#else /* USE_CHDIR_FALLBACK */
#undef  NEED_CHDIR
#define NEED_CHDIR
ssize_t
our(readlink)(const char *name, char *buf, size_t count) {
	Xl(real(readlink), name, buf, count);
}
#endif /* USE_CHDIR_FALLBACK */
#endif /* HAVE_READLINKAT */


/** CHOWN **/

int
our(fchownat)(int fd, const char *name, uid_t uid, gid_t gid, int flags) {
	Xatl(real(fchownat), fd, name, uid, gid, flags);
}

int
our(chown)(const char *name, uid_t uid, gid_t gid) {
	return our(fchownat)(AT_FDCWD, name, uid, gid, 0);
}
int
our(lchown)(const char *name, uid_t uid, gid_t gid) {
	return our(fchownat)(AT_FDCWD, name, uid, gid, AT_SYMLINK_NOFOLLOW);
}

/** CHMOD **/
int
our(chmod)(const char *path, mode_t mode) {
	int fd;
	Xfl(path, 0, -1,
	    fd, fchmod, mode);
	return fd;
}

/* MKNOD */

#ifdef HAVE_MKNODAT
int
our(mknodat)(int fd, const char *name, mode_t mode, dev_t dev) {
	Xatl(real(mknodat), fd, name, mode, dev);
}
int
our(mknod)(const char *name, mode_t mode, dev_t dev) {
	return our(mknodat)(AT_FDCWD, name, mode, dev);
}
#else /* HAVE_MKNODAT */
#ifndef USE_CHDIR_FALLBACK
#warning mknodat missing - link* calls on long paths may fail
int
our(mknod)(const char *name, mode_t mode, dev_t dev) {
	return real(mknod)(name, mode, dev);
}
#else /* USE_CHDIR_FALLBACK */
#undef  NEED_CHDIR
#define NEED_CHDIR
int
our(mknod)(const char *name, mode_t mode, dev_t dev) {
	Xl(real(mknod), name, mode, dev);
}
#endif /* USE_CHDIR_FALLBACK */
#endif /* HAVE_MKNODAT */

/* MKFIFO */

#ifdef HAVE_MKFIFOAT
int
our(mkfifoat)(int fd, const char *name, mode_t mode) {
	Xatl(real(mkfifoat), fd, name, mode);
}
int
our(mkfifo)(const char *name, mode_t mode) {
	return our(mkfifoat)(AT_FDCWD, name, mode);
}
#else /* HAVE_MKFIFOAT */
#ifndef USE_CHDIR_FALLBACK
#warning mkfifoat missing - link* calls on long paths may fail
int
our(mkfifo)(const char *name, mode_t mode) {
	return real(mkfifo)(name, mode);
}
#else /* USE_CHDIR_FALLBACK */
#undef  NEED_CHDIR
#define NEED_CHDIR
int
our(mkfifo)(const char *name, mode_t mode) {
	Xl(real(mkfifo), name, mode);
}
#endif /* USE_CHDIR_FALLBACK */
#endif /* HAVE_MKFIFOAT */


/* STAT */
int
our(fstatat)(int fd, const char *name, struct stat *sb, int flags) {
	Xatl(real(fstatat), fd, name, sb, flags);
}

int
our(stat)(const char *name, struct stat *sb) {
	return our(fstatat)(AT_FDCWD, name, sb, 0);
}
#ifdef HAVE__XSTAT
int
our(_xstat)(int ver, const char *name, struct stat *sb) {
	if (ver != _STAT_VER) {
		errno = EINVAL;
		return -1;
	}
	return our(fstatat)(AT_FDCWD, name, sb, 0);
}
#endif

int
our(lstat)(const char *name, struct stat *sb) {
	return our(fstatat)(AT_FDCWD, name, sb, AT_SYMLINK_NOFOLLOW);
}
#ifdef HAVE__XSTAT
int
our(_lxstat)(int ver, const char *name, struct stat *sb) {
	if (ver != _STAT_VER) {
		errno = EINVAL;
		return -1;
	}
	return our(fstatat)(AT_FDCWD, name, sb, AT_SYMLINK_NOFOLLOW);
}
#endif


#if !defined(_LP64)
int
our64(fstatat)(int fd, const char *name, struct stat64 *sb, int flags) {
	Xatl(real64(fstatat), fd, name, sb, flags);
}

int
our64(stat)(const char *name, struct stat64 *sb) {
	return our64(fstatat)(AT_FDCWD, name, sb, 0);
}

int
our64(lstat)(const char *name, struct stat64 *sb) {
	return our64(fstatat)(AT_FDCWD, name, sb, AT_SYMLINK_NOFOLLOW);
}
#endif

/* RMDIR */

int
our(rmdir)(const char *name) {
	return our(unlinkat)(AT_FDCWD, name, AT_REMOVEDIR);
}

/* MKDIR */

#ifdef HAVE_MKDIRAT
int
our(mkdirat)(int fd, const char *name, mode_t mode) {
	Xatl(real(mkdirat), fd, name, mode);
}

int
our(mkdir)(const char *name,  mode_t mode) {
	return our(mkdirat)(AT_FDCWD, name, mode);
}
#else /* HAVE_MKDIRAT */
#ifndef USE_CHDIR_FALLBACK
#warning mkdirat missing - link* calls on long paths may fail
int
our(mkdir)(const char *name,  mode_t mode) {
	return real(mkdir)(name, mode);
}
#else /* USE_CHDIR_FALLBACK */
#undef  NEED_CHDIR
#define NEED_CHDIR
int
our(mkdir)(const char *name,  mode_t mode) {
	Xl(real(mkdir), name, mode);
}
#endif /* USE_CHDIR_FALLBACK */
#endif /* HAVE_MKDIRAT */

/* RENAME */

int
our(renameat)(int fd1, const char *p1, int fd2, const char *p2) {
	Xatl2(real(renameat), fd1, p1, fd2, p2);
}

int
our(rename)(const char *p1, const char *p2) {
	return our(renameat)(AT_FDCWD, p1, AT_FDCWD, p2);
}


/*
 * MK*TEMP*
 *
 * there is no mk*tempat, so we need to implement it outselves
 */

/*
 * common code for all the mk*temp* - structure taken from glibc, which is
 * Copyright (C) 1991-2001, 2006, 2007, 2009 Free Software Foundation, Inc.
 */

enum temp_kind {
	T_NOCREATE,
	T_FILE,
	T_DIR
#if !defined(_LP64)
	,T_FILE64
#endif
};

/* 
 * this is for the oatl_im call: I don't see how to get around ignoring the
 * warning without having an extra version of oatl_im for (char **)
 */
GCC_DIAG_OFF("-Wcast-qual")

/*
 * return val:
 * -1 or valid fd
 *
 * AT_FDCWD is returned to signal success for T_DIR and T_NOCREATE
 */

static inline int
_mkxtempl(char *template, int slen, int flags, enum temp_kind kind) {
	size_t			len;
	int			ifd, fd;
	char     		*rest;
	char			*save, *r, *spos;
#ifdef _LP64
	struct stat		tstat;
#else
	struct stat64		tstat;
#endif

	if (slen < 0 || template == NULL || *template == '\0') {
		errno = EINVAL;
		return -1;
	}

	len = strlen(template);
	spos = NULL;
	save = NULL;

	/* check input */
	{
		char	*end, *tmp;

		end = &template[len - 1];
		/* no trailing slashes */
		if (*end == '/') {
			errno = EINVAL;
			return -1;
		}
		tmp = end;
		while (--tmp >= template) {
			if (*tmp == '/') {
				/*
				 * if we have a slash, the suffix must be within
				 * the last path element
				 */
				if (slen >= (end - tmp)) {
					errno = EINVAL;
					return -1;
				}
			}
		}
		tmp = end - slen;
		if (*tmp == 'X')
			spos = tmp + 1;
	}

	ifd = oatl_im(AT_FDCWD, template, (const char **)&rest);
	if (ifd == -1)
		return -1;

	// save = strdupa(rest);
	save = malloc(strlen(rest) + 1);
	if (! save) {
		errno = ENOMEM;
		goto err;
	}
	strcpy(save, rest);

	do {
		/*
		 * XXX: as we don't have mktempat, mktemp will check the wrong
		 * directory for long paths. this does not really matter because
		 * we check for exclusive createion / existance afterwards, but
		 * is less efficient.
		 */
		if (spos) {
			char	s = *spos;

			/* temp. trim of the suffix for mktemp */
			*spos = '\0';
			r = mktemp(rest);
			*spos = s;

			if (*r == '\0')
				goto err;
		}

		switch (kind) {
		case T_DIR:
		{
#ifdef HAVE_MKDIRAT
			if (real(mkdirat)(ifd, rest, 0700) == 0) {
				fd = AT_FDCWD;
				goto out;
			}
#else /* HAVE_MKDIRAT */
#ifndef USE_CHDIR_FALLBACK
#warning mkdirat missing - mkdtemp* calls on long paths may fail
			if (real(mkdtemp)(template)) {
				fd = AT_FDCWD;
				goto out;
			}
			goto err;
#else /* USE_CHDIR_FALLBACK */
			if (ifd == AT_FDCWD) {
				if (real(mkdir)(rest, 0700) == 0) {
					fd = AT_FDCWD;
					goto out;
				}
			} else {
				int ocwd;
				ocwd = real(open)(".", O_RDONLY|O_NDELAY|O_LARGEFILE);
				if (ocwd == -1)
					goto err;
				if (fchdir(ifd) == -1)
					goto err;
				if (real(mkdir)(rest, 0700) == 0) {
					restore_cwd(&ocwd);
					fd = AT_FDCWD;
					goto out;
				}
				restore_cwd(&ocwd);
			}
#endif /* USE_CHDIR_FALLBACK */
#endif /* HAVE_MKDIRAT */

			break; /* retry */
		}
		case T_FILE:
			fd = real(openat)(ifd, rest, O_CREAT|O_EXCL|O_RDWR|flags, 0600);

			if (fd != -1)
				goto out;

			break; /* retry */
#if !defined(_LP64)
		case T_FILE64:
			fd = real64(openat)(ifd, rest, O_CREAT|O_EXCL|O_RDWR|flags, 0600);

			if (fd != -1)
				goto out;

			break; /* retry */
#endif
		case T_NOCREATE:
		{
			int serrno = errno;
			if (our64(fstatat)(ifd, rest, &tstat, AT_SYMLINK_NOFOLLOW) == -1) {
				if (errno != ENOENT)
					goto err;

				assert(errno == ENOENT);
				errno = serrno;
				fd = AT_FDCWD;
				goto out;
			}
			break; /* retry */
		}
		default:
			assert(! "invalid kind in _mkxtempl");
		}

		if ((! spos) || (errno != EEXIST))
			goto err;
		(void) strcpy(rest, save);
	} while (1);

	assert (! "NOTREACHED");

  err:
	fd = -1;
  out:
	if (save)
		free(save);
	if (ifd != (int)AT_FDCWD)
		close_ifd(ifd);
	return fd;
}
GCC_DIAG_ON("-Wcast-qual")

/*
 * Quote solaris man page:
 *   The mkdtemp() function can set errno to the same  values  as
 *   mkdir(2).
 * so we don't need to save errno
 */

char *
our(mktemp)(char *template) {
	int serrno = errno;
	if (_mkxtempl(template, 0, 0, T_NOCREATE) == -1)
		return NULL;
	errno = serrno;
	return template;
}

char *
our(mkdtemp)(char *template) {
	if (_mkxtempl(template, 0, 0, T_DIR) == -1)
		return NULL;
	return template;
}

int
our(mkstemp)(char *template) {
	return _mkxtempl(template, 0, 0, T_FILE);
}

int
our(mkstemps)(char *template, int slen) {
	return _mkxtempl(template, slen, 0, T_FILE);
}

#if !defined(_LP64)
int
our64(mkstemp)(char *template) {
	return _mkxtempl(template, 0, 0, T_FILE64);
}

int
our64(mkstemps)(char *template, int slen) {
	return _mkxtempl(template, slen, 0, T_FILE64);
}
#endif /* ! _LP64 */

/* mkstemp GNU */
#if defined(_GNU_SOURCE)
#if defined(HAVE_MKOSTEMP)
int
our(mkostemp)(char *template, int flags) {
	return _mkxtempl(template, 0, flags, T_FILE);
}
#endif
#if defined(HAVE_MKOSTEMPS)
int
our(mkostemps)(char *template, int slen, int flags) {
	return _mkxtempl(template, slen, flags, T_FILE);
}
#endif
#if !defined(_LP64)
#if defined(HAVE_MKOSTEMP)
int
our64(mkostemp)(char *template, int flags) {
	return _mkxtempl(template, 0, flags, T_FILE64);
}
#endif
#if defined(HAVE_MKOSTEMPS)
int
our64(mkostemps)(char *template, int slen, int flags) {
	return _mkxtempl(template, slen, flags, T_FILE64);
}
#endif /* HAVE_MKOSTEMPS */
#endif /* ! _LP64 */
#endif /* _GNU_SOURCE */



/* UTIMES */

/*
 * (f)utimesat is special and strange, so map them all to utimensat, which is
 * the new syscall anyway
 */

/*
 * http://pubs.opengroup.org/onlinepubs/9699919799/functions/futimens.html#tag_16_181
 *
 * tv_nsec UTIME_NOW : same as NULL tv
 * tv_nsec UTIME_OMIT: don't set
 */

int
our(utimensat)(int fd, const char *path, const struct timespec times[2], int flag) {
	Xatl(real(utimensat), fd, path, times, flag);
}


static inline void
tv2ts(struct timespec **tsp, const struct timeval tv[2]) {
	if (tv == NULL) {
		*tsp = NULL;
	} else {
		struct timespec *ts = *tsp;
		assert(ts);
		ts[0].tv_sec  = tv[0].tv_sec;
		ts[0].tv_nsec = tv[0].tv_usec * 1000;
		ts[1].tv_sec  = tv[1].tv_sec;
		ts[1].tv_nsec = tv[1].tv_usec * 1000;
	}
}

/*
 * futimesat is special
 * - if PATH is NULL, update is applied to the fd
 * - if PATH is relative, PATH is relative to fd
 *
 * but because utimensatl does not handle PATH == NULL, we need
 * to pass to futimens
 *
 */
int
our(futimesat)(int filedes, const char *path, const struct timeval tv[2]) {
	struct timespec ts[2];
	struct timespec *tsp = ts;
	
	tv2ts(&tsp, tv);

	if (path == NULL)
		return futimens(filedes, tsp);

	return our(utimensat)(filedes, path, tsp, 0);
}

int
our(utimes)(const char *path, const struct timeval tv[2]) {
	struct timespec ts[2];
	struct timespec *tsp = ts;

	tv2ts(&tsp, tv);

	return our(utimensat)(AT_FDCWD, path, tsp, 0);
}

#if defined(HAVE_LUTIMES)
int
our(lutimes)(const char *path, const struct timeval tv[2]) {
	struct timespec ts[2];
	struct timespec *tsp = ts;

	tv2ts(&tsp, tv);

	return our(utimensat)(AT_FDCWD, path, tsp, AT_SYMLINK_NOFOLLOW);
}
#endif

int
our(utime)(const char *path, const struct utimbuf *times) {
	if (times) {
		struct timespec ts[2];

		ts[0].tv_sec  = times->actime;
		ts[0].tv_nsec = 0;

		ts[1].tv_sec  = times->modtime;
		ts[1].tv_nsec = 0;

		return our(utimensat)(AT_FDCWD, path, ts, 0);
	}

	return our(utimensat)(AT_FDCWD, path, NULL, 0);
}

/* OPENDIR */
DIR *
our(opendir)(const char *dirname) {
	int fd;
	/* almost identical to Xfl, but we must not close the fd */
	fd = _openat(AT_FDCWD, dirname, O_RDONLY|O_NDELAY|O_LARGEFILE, 0);

	if (fd == -1)
		return NULL;

	return fdopendir(fd);
}

#ifdef NEED_CHDIR

/*
 * this function will always try to chdir even if path is shorter than PATH_MAX
 * so it can be used on a common prefix of two paths
 */
static int
chdir_im(const char *path, const char **remaining, int limit) {
	char		interm[PATH_MAX];
	const char	*p, *sl;
	size_t		l;

	/*
	 * work with functions on NULL paths like fchownat
	 * - at any rate we can leave error handling with the system function
	 */
	if (path == NULL) {
		*remaining = NULL;
		return 0;
	}

	l = strlen(path);
	if ((limit > 0) && (l > limit))
		l = limit;

	p = path;
	do {
		/* find last slash within limit */
		sl = p + ((l < PATH_MAX) ? l : (PATH_MAX - 1));

		while (sl > p) {
			if (*sl == '/')
				goto found;
			sl--;
		}

		goto out;

	  found:
		assert((sl - p) < PATH_MAX);
		strncpy(interm, p, (sl - p));
		interm[sl - p] = '\0';

		if (chdir(interm) == -1)
			goto err;

		l -= (sl - p) + 1;
		p  = sl + 1;
	} while (1);

  out:
	*remaining = p;
	return 0;

  err:
	return -1;
}

static inline void
restore_cwd(int *fd) {
	assert(*fd >= 0);
	assert(*fd != (int)AT_FDCWD);

	if (fchdir(*fd) == -1) {
		/*
		 * XXX: This really must not happen, we have
		 *      changed the processes' state and can't
		 *      restore it. Any better ideas?
		 */
		assert(! "liblongpath restore_cwd");
		perror("liblongpath restore_cwd");
		exit(1);
	}

	close_ifd(*fd);
	*fd = -1;
}

#else /* NEED_CHDIR */

static int chdir_im(const char *path, const char **remaining, int limit) {
	assert(! "not used");
	(void)path;
	(void)remaining;
	(void)limit;
	return -1;
}
static inline void restore_cwd(int *fd) {
	assert(! "not used");
	(void)fd;
}
int  (* _dummy1)(const char *, const char **, int) = chdir_im;
void (* _dummy2)(int *) = restore_cwd;

#endif /* NEED_CHDIR */
